from abc import ABC, abstractmethod


class DenserEmbeddings(ABC):
    embedding_size: int
    
    @abstractmethod
    def embed_documents(self, texts) -> list:
        pass

    @abstractmethod
    def embed_query(self, text) -> list:
        pass


class SentenceTransformerEmbeddings(DenserEmbeddings):
    def __init__(self, model_name: str, embedding_size: int, one_model: bool):
        try:
            import sentence_transformers
        except ImportError as exc:
            raise ImportError(
                "Could not import sentence_transformers python package. "
                "Please install it with `pip install sentence-transformers`."
            ) from exc

        self.client = sentence_transformers.SentenceTransformer(
            model_name, trust_remote_code=True
        )
        self.embedding_size = embedding_size
        self.one_model = one_model

    def embed_documents(self, texts):
        return self.client.encode(texts)

    def embed_query(self, text):
        if self.one_model:
            embeddings = self.client.encode([text])
        else:
            embeddings = self.client.encode([text], prompt_name="query")
        return embeddings


class VoyageAPIEmbeddings(DenserEmbeddings):
    def __init__(self, api_key: str, model_name: str, embedding_size: int):
        try:
            import voyageai
        except ImportError as exc:
            raise ImportError(
                "Could not import voyage python package. "
                "Please install it with `pip install voyageai`."
            ) from exc

        self.client = voyageai.Client(api_key)
        self.model_name = model_name
        self.embedding_size = embedding_size

    def embed_documents(self, texts):
        """
        Embeds multiple documents using the Voyage API.
        Args:
            texts: A list of document texts.
        Returns:
            A list of document embeddings.
        """
        embeddings = self.client.embed(texts, model=self.model_name).embeddings
        return embeddings

    def embed_query(self, text):
        """
        Embeds a single query using the Voyage API.
        Args:
            text: The query text.
        Returns:
            The query embedding.
        """
        embeddings = self.client.embed([text], model=self.model_name).embeddings
        return embeddings



from asyncio.log import logger
from typing import Any, Dict, List, Optional, Tuple
import uuid
import time

from langchain_core.documents import Document

from denser_retriever.embeddings import DenserEmbeddings
from denser_retriever.gradient_boost import DenserGradientBoost
from denser_retriever.keyword import DenserKeywordSearch
from denser_retriever.reranker import DenserReranker
from denser_retriever.utils import (
    docs_to_dict,
    merge_results,
    min_max_normalize,
    parse_features,
    scale_results,
    standardize_normalize,
)
from denser_retriever.vectordb.base import DenserVectorDB

config_to_features = {
    "es+vs": ["1,2,3,4,5,6", None],
    "es+rr": ["1,2,3,7,8,9", None],
    "vs+rr": ["4,5,6,7,8,9", None],
    "es+vs+rr": ["1,2,3,4,5,6,7,8,9", None],
    "es+vs_n": ["1,2,3,4,5,6", "2,5"],
    "es+rr_n": ["1,2,3,7,8,9", "2,8"],
    "vs+rr_n": ["4,5,6,7,8,9", "5,8"],
    "es+vs+rr_n": ["1,2,3,4,5,6,7,8,9", "2,5,8"],
}


class DenserRetriever:
    def __init__(
        self,
        index_name: str,
        keyword_search: Optional[DenserKeywordSearch],
        vector_db: Optional[DenserVectorDB],
        reranker: Optional[DenserReranker],
        embeddings: DenserEmbeddings,
        gradient_boost: Optional[DenserGradientBoost],
        combine_mode: str = "linear",
        xgb_model_features: str = "es+vs+rr_n",
        search_fields: List[str] = []
    ):
        # config parameters
        self.index_name = index_name
        self.combine_mode = combine_mode
        self.xgb_model_features = config_to_features[xgb_model_features]

        # models
        self.embeddings = embeddings
        self.gradient_boost = gradient_boost
        self.keyword_search = keyword_search
        self.vector_db = vector_db
        self.reranker = reranker

        # create index. If exists, remove them first if drop_old is true
        if self.vector_db:
            assert embeddings
            self.vector_db.create_index(index_name, embeddings, search_fields)
        if self.keyword_search:
            self.keyword_search.create_index(index_name, search_fields)

    def ingest(self, docs: List[Document], overwrite_pid: bool = True) -> List[str]:
        # add pid into metadata for each document
        if overwrite_pid:
            for _, doc in enumerate(docs):
                doc.metadata["pid"] = uuid.uuid4().hex
        if self.keyword_search:
            logger.info(f"Adding {len(docs)} documents to keyword search")
            self.keyword_search.add_documents(docs)
            logger.info(f"Done adding {len(docs)} documents to keyword search")
        if self.vector_db:
            logger.info(f"Adding {len(docs)} documents to vector db")
            self.vector_db.add_documents(documents=docs)
            logger.info(f"Done adding {len(docs)} documents to vector db")

        return [doc.metadata["pid"] for doc in docs]

    def retrieve(
        self, query: str, k: int = 100, filter: Dict[str, Any] = {}, **kwargs: Any
    ):
        logger.info(f"Retrieve query: {query} top_k: {k}")
        if self.combine_mode in ["linear", "rank"]:
            return self._retrieve_by_linear_or_rank(query, k, filter, **kwargs)
        else:
            return self._retrieve_by_model(query, k, filter, **kwargs)

    def _retrieve_by_linear_or_rank(
        self, query: str, k: int = 100, filter: Dict[str, Any] = {}, **kwargs: Any
    ):
        passages = []

        if self.keyword_search:
            es_docs = self.keyword_search.retrieve(
                query, self.keyword_search.top_k, filter=filter, **kwargs
            )
            es_passages = scale_results(es_docs, self.keyword_search.weight)
            logger.info(f"Keyword search: {len(es_passages)}")
            passages.extend(es_passages)

        if self.vector_db:
            vector_docs = self.vector_db.similarity_search_with_score(
                query, self.vector_db.top_k, filter, **kwargs
            )
            logger.info(f"Vector search: {len(vector_docs)}")
            passages = merge_results(
                passages, vector_docs, 1.0, self.vector_db.weight, self.combine_mode
            )

        if self.reranker:
            start_time = time.time()
            docs = [doc for doc, _ in passages[: self.reranker.top_k]]
            reranked_docs = self.reranker.rerank(docs, query)

            passages = merge_results(
                passages,
                reranked_docs,
                1.0,
                self.reranker.weight,
                self.combine_mode,
            )
            rerank_time_sec = time.time() - start_time
            logger.info(f"Rerank time: {rerank_time_sec:.3f} sec.")

        return passages[:k]

    def _retrieve_by_model(
        self, query: str, k: int = 100, filter: Dict[str, Any] = {}, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        docs, doc_features = self._retrieve_with_features(query, filter, **kwargs)

        if not self.gradient_boost:
            raise ValueError("Gradient Boost model not provided")

        csr_data = parse_features(doc_features)
        pred = self.gradient_boost.predict(csr_data)

        assert len(pred) == len(docs)
        scores = pred.tolist()
        logger.info(f"xgb prediction scores: {scores}")
        reranked_docs = list(zip(docs, scores))
        reranked_docs.sort(key=lambda x: x[1], reverse=True)

        return reranked_docs[:k]

    def _retrieve_with_features(
        self, query: str, filter: Dict[str, Any] = {}, **kwargs: Any
    ) -> Tuple[List[Document], List[List[str]]]:
        ks_docs = []
        if self.keyword_search:
            ks_docs = self.keyword_search.retrieve(
                query, self.keyword_search.top_k, filter=filter, **kwargs
            )
        vs_docs = []
        if self.vector_db:
            vs_docs = self.vector_db.similarity_search_with_score(
                query, k=self.vector_db.top_k, filter=filter, **kwargs
            )

        combined = []
        seen = set()
        for item in ks_docs + vs_docs:
            if item[0].metadata["pid"] not in seen:
                combined.append(item)
                seen.add(item[0].metadata["pid"])

        combined_docs = [doc for doc, _ in combined]

        reranked_docs = []
        if self.reranker:
            reranked_docs = self.reranker.rerank(combined_docs, query)

        _, ks_score_dict, ks_rank_dict = docs_to_dict(ks_docs)
        _, vs_score_dict, vs_rank_dict = docs_to_dict(vs_docs)
        reranked_docs_dict, reranked_score_dict, reranked_rank_dict = docs_to_dict(
            reranked_docs
        )

        docs, doc_features = [], []
        for pid in reranked_docs_dict.keys():
            docs.append(reranked_docs_dict[pid])

            features = []
            features.append(0)  # placeholder
            features.append(ks_rank_dict.get(pid, -1))  # 1. keyword rank
            features.append(ks_score_dict.get(pid, -1000))  # 2. keyword score
            miss = 1 if ks_rank_dict.get(pid, -1) == -1 else 0
            features.append(miss)  # 3. keyword miss

            features.append(vs_rank_dict.get(pid, -1))  # 4. vector rank
            features.append(vs_score_dict.get(pid, -1000))  # 5. vector score
            miss = 1 if vs_rank_dict.get(pid, -1) == -1 else 0
            features.append(miss)  # 6. vector miss

            assert pid in reranked_rank_dict
            features.append(reranked_rank_dict[pid])  # 7. rerank rank
            features.append(reranked_score_dict[pid])  # 8. rerank score
            features.append(0)  # 9. placeholder
            doc_features.append(features)

        features_to_use, features_to_normalize = self.xgb_model_features
        features_to_use = features_to_use.split(",")
        features_to_normalize = features_to_normalize.split(",")

        if features_to_normalize:
            features_raw = {f: [] for f in features_to_normalize}
            for data in doc_features:
                for f_name in features_to_normalize:
                    features_raw[f_name].append(float(data[int(f_name)]))

            # normalize features_raw
            standardized_features = {}
            min_max_features = {}
            for f_name in features_to_normalize:
                standardized_features[f_name] = standardize_normalize(
                    features_raw[f_name]
                )
                min_max_features[f_name] = min_max_normalize(features_raw[f_name])

        non_zero_normalized_features = []
        for i, data in enumerate(doc_features):
            features = []
            for f_id in features_to_use:
                f_value = data[int(f_id)]
                if f_value != 0.0:
                    features.append(f"{f_id}:{f_value}")

            if features_to_normalize:
                f_id = len(data[1:]) + 1
                normalized_features = []
                for j, f in enumerate(features_to_normalize):
                    if standardized_features[f][i] != 0.0:
                        normalized_features.append(
                            f"{f_id + 2 * j}:{standardized_features[f][i]}"
                        )
                    if min_max_features[f][i] != 0.0:
                        normalized_features.append(
                            f"{f_id + 2 * j + 1}:{min_max_features[f][i]}"
                        )
                non_zero_normalized_features.append(
                    [data[0]] + features + normalized_features
                )
            else:
                non_zero_normalized_features.append([data[0]] + features)

        return docs, non_zero_normalized_features

    def delete(self, ids: Optional[List[str]] = None, source_id: Optional[str] = None, **kwargs: str):
        """Clear the retriever."""
        if self.vector_db:
            self.vector_db.delete(ids=ids, source_id=source_id, **kwargs)
        if self.keyword_search:
            self.keyword_search.delete(ids=ids, source_id=source_id, **kwargs)

    def delete_all(self):
        """Clear the retriever."""
        if self.vector_db:
            self.vector_db.delete_all()
        if self.keyword_search:
            self.keyword_search.delete_all()

    def get_field_categories(self, field, k: int = 10):
        """
        Get the categories of a field.

        Args:
            field: The field to get the categories of.
            k: The number of categories to return.

        Returns:
            A list of categories.
        """
        if not self.keyword_search:
            raise ValueError("Keyword search not initialized")
        return self.keyword_search.get_categories(field, k)

    def get_filter_fields(self):
        """Get the filter fields."""
        if not self.keyword_search:
            raise ValueError("Keyword search not initialized")
        return self.keyword_search.get_index_mappings()



from abc import ABC, abstractmethod
from numpy import ndarray
import xgboost as xgb
from scipy.sparse import csr_matrix


class DenserGradientBoost(ABC):
    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def predict(self, csr_data: csr_matrix) -> ndarray:
        pass


class XGradientBoost(DenserGradientBoost):
    def __init__(
        self,
        xgb_model_path: str,
    ):
        self.xgb_model_path = xgb_model_path
        self.load()

    def load(self):
        self.model = xgb.Booster()
        self.model.load_model(self.xgb_model_path)

    def predict(self, csr_data: csr_matrix):
        return self.model.predict(xgb.DMatrix(csr_data))



from abc import ABC, abstractmethod
import operator
from typing import List, Sequence, Tuple
import time
import logging
import cohere
from sentence_transformers import CrossEncoder
from langchain_core.documents import Document

logger = logging.getLogger(__name__)

class DenserReranker(ABC):
    def __init__(self, top_k: int = 50, weight: float = 0.5):
        self.top_k = top_k
        self.weight = weight

    @abstractmethod
    def rerank(
        self,
        documents: Sequence[Document],
        query: str,
    ) -> List[Tuple[Document, float]]:
        pass


class HFReranker(DenserReranker):
    """Rerank documents using a HuggingFaceCrossEncoder model."""

    def __init__(self, model_name: str, top_k: int, **kwargs):
        super().__init__(top_k=top_k)
        self.model = CrossEncoder(model_name, **kwargs)

    def rerank(
        self,
        documents: Sequence[Document],
        query: str,
    ) -> List[Tuple[Document, float]]:
        """
        Rerank documents using CrossEncoder.

        Args:
            documents: A sequence of documents to rerank.
            query: The query to use for ranking the documents.

        Returns:
            A list of tuples containing the document and its score.
        """
        if not documents:
            return []
        start_time = time.time()
        scores = self.model.predict([(query, doc.page_content) for doc in documents], convert_to_tensor=True)
        docs_with_scores = list(zip(documents, scores))
        result = sorted(docs_with_scores, key=operator.itemgetter(1), reverse=True)
        rerank_time_sec = time.time() - start_time
        logger.info(f"Rerank time: {rerank_time_sec:.3f} sec.")
        logger.info(f"Reranked {len(result)} documents.")
        return result


class CohereReranker(DenserReranker):
    """Rerank documents using the Cohere API."""

    def __init__(self, api_key: str, model_name: str = "rerank-english-v3.0", **kwargs):
        """
        Initialize Cohere reranker.

        Args:
            api_key: The API key for Cohere.
            model_name: The name of the Cohere model to use for reranking.
        """
        super().__init__()
        self.client = cohere.Client(api_key)
        self.model_name = model_name

    def rerank(
            self,
            documents: Sequence[Document],
            query: str,
    ) -> List[Tuple[Document, float]]:
        """
        Rerank documents using Cohere's reranking model.

        Args:
            documents: A sequence of documents to rerank.
            query: The query to use for ranking the documents.

        Returns:
            A list of tuples containing the document and its score.
        """
        if not documents:
            return []

        start_time = time.time()

        # Prepare documents for reranking
        texts = [doc.page_content for doc in documents]
        response = self.client.rerank(
            model=self.model_name,
            query=query,
            documents=texts
        )
        # Combine documents with scores from the rerank response
        docs_with_scores = [(documents[result.index], result.relevance_score) for result in response.results]

        # Sort the documents by their scores in descending order
        result = sorted(docs_with_scores, key=operator.itemgetter(1), reverse=True)

        rerank_time_sec = time.time() - start_time
        logger.info(f"Cohere Rerank time: {rerank_time_sec:.3f} sec.")
        logger.info(f"Reranked {len(result)} documents.")
        return result



import json
from typing import Dict, List, Optional, Tuple

import numpy as np
import pytrec_eval

from scipy.sparse import csr_matrix
from collections import defaultdict
from langchain_core.documents import Document

def evaluate(
    qrels: Dict[str, Dict[str, int]],
    results: Dict[str, Dict[str, float]],
    metric_file: Optional[str] = None,
    k_values: List[int] = [1, 3, 5, 10, 20, 100, 1000],
    ignore_identical_ids: bool = True,
) -> Tuple[Dict[str, float], Dict[str, float], Dict[str, float], Dict[str, float]]:
    if ignore_identical_ids:
        print(
            "For evaluation, we ignore identical query and document ids (default), please explicitly set ``ignore_identical_ids=False`` to ignore this."  # noqa: E501
        )
        popped = []
        for qid, rels in results.items():
            for pid in list(rels):
                if qid == pid:
                    results[qid].pop(pid)
                    popped.append(pid)

    ndcg = {}
    _map = {}
    recall = {}
    precision = {}

    for k in k_values:
        ndcg[f"NDCG@{k}"] = 0.0
        _map[f"MAP@{k}"] = 0.0
        recall[f"Recall@{k}"] = 0.0
        precision[f"P@{k}"] = 0.0

    map_string = "map_cut." + ",".join([str(k) for k in k_values])
    ndcg_string = "ndcg_cut." + ",".join([str(k) for k in k_values])
    recall_string = "recall." + ",".join([str(k) for k in k_values])
    precision_string = "P." + ",".join([str(k) for k in k_values])
    evaluator = pytrec_eval.RelevanceEvaluator(
        qrels, {map_string, ndcg_string, recall_string, precision_string}
    )
    scores = evaluator.evaluate(results)

    for query_id in scores.keys():
        for k in k_values:
            ndcg[f"NDCG@{k}"] += scores[query_id]["ndcg_cut_" + str(k)]
            _map[f"MAP@{k}"] += scores[query_id]["map_cut_" + str(k)]
            recall[f"Recall@{k}"] += scores[query_id]["recall_" + str(k)]
            precision[f"P@{k}"] += scores[query_id]["P_" + str(k)]

    for k in k_values:
        ndcg[f"NDCG@{k}"] = round(ndcg[f"NDCG@{k}"] / len(scores), 5)
        _map[f"MAP@{k}"] = round(_map[f"MAP@{k}"] / len(scores), 5)
        recall[f"Recall@{k}"] = round(recall[f"Recall@{k}"] / len(scores), 5)
        precision[f"P@{k}"] = round(precision[f"P@{k}"] / len(scores), 5)

    if metric_file:
        out = open(metric_file, "w")
        for eval in [ndcg, _map, recall, precision]:
            json.dump(eval, out, indent=4, ensure_ascii=False)
            out.write("\n")

    return ndcg, _map, recall, precision



def save_queries(queries, output_file: str):
    out = open(output_file, "w")
    for d in queries:
        data = {"id": d["id"], "text": d["text"]}
        json.dump(data, out, ensure_ascii=False)
        out.write("\n")

def load_queries(in_file: str):
    res = []
    for line in open(in_file, "r"):
        res.append(json.loads(line))
    return res

def save_qrels(qrels, output_file: str):
    out = open(output_file, "w")
    for q in qrels.keys():
        data = {q: qrels[q]}
        json.dump(data, out, ensure_ascii=False)
        out.write("\n")

def save_qrels_from_trec(trec_file, qrels_file):
    qrels = {}
    with open(trec_file, "r") as f:
        for line in f:
            qid, _, pid, rel = line.split()
            if qid not in qrels:
                qrels[qid] = {}
            qrels[qid][pid] = int(rel)

    save_qrels(qrels, qrels_file)
    return qrels

def load_qrels(in_file: str):
    res = {}
    for line in open(in_file, "r"):
        obj = json.loads(line)
        assert len(list(obj.keys())) == 1
        query = list(obj.keys())[0]
        res[query] = obj[query]
    return res

def passages_to_dict(passages: List[Dict[str, str]], doc_task, score_name="score"):
    res = {}
    for passage in passages:
        source, pid = passage["source"], passage.get("pid", -1)
        if doc_task or pid == -1:
            uid_str = source
        else:
            uid_str = f"{source}-{pid}"
        assert uid_str not in res
        res[uid_str] = passage[score_name]
    return res



def standardize_normalize(data):
    # Convert list to numpy array for easier manipulation
    arr = np.array(data)

    # Calculate mean and standard deviation of the array
    mean = np.mean(arr)
    std_dev = np.std(arr)

    # Standardize the array
    if std_dev != 0:
        standardized_arr = (arr - mean) / std_dev
    else:
        standardized_arr = arr - mean
        # print("Warning: std_dev is zero")

    # Convert back to list if necessary
    return standardized_arr.tolist()


def min_max_normalize(data):
    # Convert list to numpy array for easier manipulation
    arr = np.array(data)

    # Find the minimum and maximum values of the array
    min_val = np.min(arr)
    max_val = np.max(arr)

    # Apply min-max normalization
    if max_val - min_val != 0:
        normalized_arr = (arr - min_val) / (max_val - min_val)
    else:
        normalized_arr = arr - min_val
        # print("Warning: max_val - min_val is zero")

    # Convert back to list if needed
    return normalized_arr.tolist()


# Parse the features
def parse_features(features):
    row = []
    col = []
    value = []

    for i, sample in enumerate(features):
        for feature in sample[1:]:  # Skip the first element as it is the label
            feature_id, feature_value = feature.split(":")
            row.append(i)
            col.append(int(feature_id) - 1)  # We have feature_id starting from 1
            value.append(float(feature_value))

    num_samples = len(features)
    num_features = max(col) + 1  # Determine the number of features

    return csr_matrix((value, (row, col)), shape=(num_samples, num_features))


def scale_results(passages: List[Tuple[Document, float]], weight: float):
    return [(doc, score * weight) for doc, score in passages]


def merge_score_linear(
        passages_1: List[Tuple[Document, float]],
        passages_2: List[Tuple[Document, float]],
        weight_1: float,
        weight_2: float,
) -> List[Tuple[Document, float]]:
    # Dictionary to store the combined scores with pid as the key
    combined_scores: Dict[str, float] = defaultdict(float)
    document_map: Dict[str, Document] = {}

    # Add the weighted scores from passages_1
    for doc, score in passages_1:
        pid = doc.metadata["pid"]
        combined_scores[pid] += score * weight_1
        document_map[pid] = doc

    # Add the weighted scores from passages_2
    for doc, score in passages_2:
        pid = doc.metadata["pid"]
        combined_scores[pid] += score * weight_2
        document_map[pid] = doc

    # Create a list of tuples from the combined scores and sort by score in descending order
    merged_passages = [(document_map[pid], score) for pid, score in combined_scores.items()]
    merged_passages.sort(key=lambda x: x[1], reverse=True)

    return merged_passages


# TODO: to fix this function
def merge_score_rank(
    passages_1: List[Tuple[Document, float]], passages_2: List[Tuple[Document, float]]
) -> List[Tuple[Document, float]]:
    rank_dict: Dict[int, float] = {}
    doc_dict: Dict[int, Document] = {}
    k = 60

    for rank, (document, _) in enumerate(passages_1, start=1):
        rank_dict[rank] = 1 / (k + rank)
        doc_dict[rank] = document

    offset = len(passages_1)
    for rank, (document, _) in enumerate(passages_2, start=1):
        rank_dict[offset + rank] = rank_dict.get(offset + rank, 0) + 1 / (k + rank)
        doc_dict[offset + rank] = document

    return [(doc_dict[idx], score) for idx, score in rank_dict.items()]


def merge_results(
    passages_1: List[Tuple[Document, float]],
    passages_2: List[Tuple[Document, float]],
    weight_1: float,
    weight_2: float,
    combine: str,
) -> List[Tuple[Document, float]]:
    if combine == "linear":
        merged_passages = merge_score_linear(passages_1, passages_2, weight_1, weight_2)
    else:  # rank
        merged_passages = merge_score_rank(passages_1, passages_2)

    return merged_passages


def docs_to_dict(
    doc: List[Tuple[Document, float]],
) -> Tuple[Dict[str, Document], Dict[str, float], Dict[str, int]]:
    """Convert a list of documents and scores to dictionaries."""
    doc_dict, score_dict, rank_dict = {}, {}, {}

    for i, (document, score) in enumerate(doc):
        uid_str = document.metadata.get("pid")
        # store the document, score and rank
        doc_dict[uid_str] = document
        score_dict[uid_str] = score
        rank_dict[uid_str] = i + 1

    return doc_dict, score_dict, rank_dict



# type: ignore[attr-defined]
"""Enterprise-grade AI retriever solution that seamlessly integrates to enhance your AI applications."""

import sys
from .embeddings import (
    DenserEmbeddings,
    SentenceTransformerEmbeddings,
    VoyageAPIEmbeddings,
)
from .keyword import DenserKeywordSearch, ElasticKeywordSearch
from .reranker import DenserReranker, HFReranker, CohereReranker
from .retriever import DenserRetriever
from .vectordb.base import DenserVectorDB
from .vectordb.milvus import MilvusDenserVectorDB

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()

__all__ = [
    "DenserEmbeddings",
    "SentenceTransformerEmbeddings",
    "VoyageAPIEmbeddings",
    "DenserKeywordSearch",
    "ElasticKeywordSearch",
    "DenserReranker",
    "HFReranker",
    "CohereReranker",
    "DenserRetriever",
    "DenserVectorDB",
    "MilvusDenserVectorDB",
]



from abc import ABC
from typing import Any, Dict, List, Optional, Tuple

from langchain_core.documents import Document
from denser_retriever.embeddings import DenserEmbeddings


class DenserVectorDB(ABC):
    """
    Interface for a denser vector database.
    """

    def __init__(self, top_k: int = 100, weight: float = 0.5):
        self.top_k = top_k
        self.weight = weight

    def create_index(
        self,
        index_name: str,
        embeddings: DenserEmbeddings,
        search_fields: List[str],
        **args: Any,
    ):
        raise NotImplementedError(
            f"create_index has not been implemented for {self.__class__.__name__}"
        )

    def add_documents(
        self,
        documents: List[Document],
        **kwargs: Any,
    ) -> List[str]:
        raise NotImplementedError(
            f"upsert has not been implemented for {self.__class__.__name__}"
        )

    def similarity_search_with_score(
        self,
        query: str,
        k: int = 100,
        filter: Dict[str, Any] = {},
        **kwargs: Any,
    ) -> List[Tuple[Document, float]]:
        raise NotImplementedError(
            f"similarity_search_with_score has not been implemented for {self.__class__.__name__}"
        )

    def filter_expression(
        self,
        filter_dict: Dict[str, Any],
    ) -> Any:
        raise NotImplementedError(
            f"filter_expression has not been implemented for {self.__class__.__name__}"
        )

    def delete(
        self,
        ids: Optional[List[str]] = None,
        source_id: Optional[str] = None,
        source_url: Optional[str] = None,
        **kwargs: str,
    ):
        raise NotImplementedError(
            f"clear has not been implemented for {self.__class__.__name__}"
        )

    def delete_all(self):
        raise NotImplementedError(
            f"clear has not been implemented for {self.__class__.__name__}"
        )



import time
from datetime import datetime
from typing import Any, Dict


def generate_milvus_expr(filter_dict: Dict[str, Any]) -> str:
    """Generate a Milvus expression from a filter dictionary."""
    expressions = []
    for key, value in filter_dict.items():
        if value is None:
            continue
        if isinstance(value, tuple) and len(value) == 2:
            start, end = value
            expressions.append(f"{key} >= '{start}' and {key} <= '{end}'")
        else:
            expressions.append(f"{key} == '{value}'")
    return " and ".join(expressions)


class FieldMapper:
    def __init__(self, fields: list[str]):
        self.original_to_key_map: Dict[str, str] = {}
        self.key_to_original_map: Dict[str, str] = {}
        self.field_types: Dict[str, str] = {}
        self.category_to_number_map = {}
        self.number_to_category_map = {}

        for f in fields:
            comps = f.split(":")
            assert len(comps) == 2 or len(comps) == 3

            if len(comps) == 3:
                original_key = comps[0]
                key = comps[1]
                field_type = comps[2]
            elif len(comps) == 2:
                original_key = key = comps[0]
                field_type = comps[1]

            self.original_to_key_map[original_key] = key
            self.key_to_original_map[key] = original_key
            self.field_types[key] = field_type

            if field_type == 'keyword':
                self.category_to_number_map[key] = {}
                self.number_to_category_map[key] = {}

    def get_key(self, original_key):
        return self.original_to_key_map.get(original_key)

    def get_keys(self) -> list[str]:
        return list(self.original_to_key_map.values())

    def get_original_key(self, key):
        return self.key_to_original_map.get(key)

    def get_original_keys(self) -> list[str]:
        return list(self.key_to_original_map.values())

    def get_field_type(self, key) -> str | None:
        return self.field_types.get(key)

    def convert_for_storage(self, data: Dict[str, Any]) -> Any:
        converted_data = {}
        for key, value in data.items():
            converted_key = self.get_key(key)
            if converted_key is None:
                continue
            converted_value = self.convert_query_condition(converted_key, value)
            converted_data[converted_key] = converted_value
        return converted_data[converted_key]

    def convert_to_original(self, data):
        original_data = {}
        for key, value in data.items():
            original_key = self.key_to_original_map.get(key)
            original_value = self.convert_back_to_original(key, value)
            original_data[original_key] = original_value
        return original_data[original_key]

    def convert_query_condition(self, key: str, value: str) -> Any:
        field_type = self.get_field_type(key)
        if field_type == "date":
            return self.convert_date_to_timestamp(value)
        elif field_type == "keyword":
            return self.convert_category_to_number(key, value)

    def convert_back_to_original(self, key, value):
        field_type = self.field_types.get(key)
        if field_type == 'date':
            return self.convert_timestamp_to_date(value)
        elif field_type == 'keyword':
            return self.convert_number_to_category(key, value)
        else:
            return value

    def convert_date_to_timestamp(self, date_str):
        dt = datetime.strptime(date_str, "%Y-%m-%d")
        timestamp = int(time.mktime(dt.timetuple()))
        return timestamp

    def convert_timestamp_to_date(self, timestamp):
        dt = datetime.fromtimestamp(timestamp)
        return dt.strftime("%Y-%m-%d")

    def convert_category_to_number(self, key, value):
        category_map = self.category_to_number_map[key]
        if value not in category_map:
            category_map[value] = len(category_map) + 1
        return category_map[value]

    def convert_number_to_category(self, key, number):
        return self.number_to_category_map[key].get(number)



from abc import ABC, abstractmethod
import logging
from typing import Any, Dict, List, Optional, Tuple
import uuid
import time

from elasticsearch import Elasticsearch

from langchain_core.documents import Document
from denser_retriever.filter import FieldMapper

logger = logging.getLogger(__name__)


def create_elasticsearch_client(
    url: Optional[str] = None,
    cloud_id: Optional[str] = None,
    api_key: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    params: Optional[Dict[str, Any]] = None,
) -> Elasticsearch:
    if url and cloud_id:
        raise ValueError(
            "Both es_url and cloud_id are defined. Please provide only one."
        )

    connection_params: Dict[str, Any] = {}

    if url:
        connection_params["hosts"] = [url]
    elif cloud_id:
        connection_params["cloud_id"] = cloud_id
    else:
        raise ValueError("Please provide either elasticsearch_url or cloud_id.")

    if api_key:
        connection_params["api_key"] = api_key
    elif username and password:
        connection_params["basic_auth"] = (username, password)

    if params is not None:
        connection_params.update(params)

    es_client = Elasticsearch(**connection_params)

    es_client.info()  # test connection

    return es_client


class DenserKeywordSearch(ABC):
    """
    Denser keyword search interface.
    """

    def __init__(self, top_k: int = 100, weight: float = 0.5):
        self.top_k = top_k
        self.weight = weight

    @abstractmethod
    def create_index(self, index_name: str, search_fields: List[str], **args: Any):
        raise NotImplementedError

    @abstractmethod
    def add_documents(
        self,
        documents: List[Document],
        **kwargs: Any,
    ) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def retrieve(
        self,
        query: str,
        k: int = 100,
        filter: Dict[str, Any] = {},
    ) -> List[Tuple[Document, float]]:
        raise NotImplementedError

    @abstractmethod
    def get_index_mappings(self) -> Dict[Any, Any]:
        raise NotImplementedError

    @abstractmethod
    def get_categories(self, field: str, k: int = 10) -> List[Any]:
        raise NotImplementedError

    @abstractmethod
    def delete(
        self,
        ids: Optional[List[str]] = None,
        source_id: Optional[str] = None,
        source_url: Optional[str] = None,
        **kwargs: str,
    ):
        raise NotImplementedError

    @abstractmethod
    def delete_all(self):
        raise NotImplementedError


class ElasticKeywordSearch(DenserKeywordSearch):
    """
    Elasticsearch keyword search class.
    """

    index_name: str
    """Index name for retrieval"""
    client: Elasticsearch
    """Elasticsearch client"""
    search_fields: FieldMapper
    """Fields to be indexed"""
    analysis: Optional[str]
    """Analysis type"""

    def __init__(
        self,
        drop_old: Optional[bool],
        es_connection: Elasticsearch,
        analysis: Optional[str] = "default",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.drop_old = drop_old
        self.analysis = analysis
        self.client = es_connection

    def create_index(self, index_name: str, search_fields: List[str], **args: Any):

        # Define the index settings and mappings
        self.index_name = index_name
        self.search_fields = FieldMapper(search_fields)

        logger.info("ES analysis %s", self.analysis)
        if self.analysis == "default":
            settings = {
                "analysis": {"analyzer": {"default": {"type": "standard"}}},
                "similarity": {
                    "custom_bm25": {
                        "type": "BM25",
                        "k1": 1.2,
                        "b": 0.75,
                    }
                },
            }
            mappings = {
                "properties": {
                    "content": {
                        "type": "text",
                        "similarity": "custom_bm25",  # Use the custom BM25 similarity
                    },
                    "title": {
                        "type": "text",
                    },
                    "source": {
                        "type": "keyword",
                    },
                    "pid": {
                        "type": "text",
                    },
                }
            }
        else:  # ik
            settings = {
                "analysis": {
                    "analyzer": {
                        "ik_max_word": {"type": "custom", "tokenizer": "ik_max_word"},
                        "ik_smart": {"type": "custom", "tokenizer": "ik_smart"},
                    }
                },
                "similarity": {
                    "custom_bm25": {
                        "type": "BM25",
                        "k1": 1.2,
                        "b": 0.75,
                    }
                },
            }
            mappings = {
                "properties": {
                    "content": {
                        "type": "text",
                        "analyzer": "ik_max_word",
                        "similarity": "custom_bm25",  # Use the custom BM25 similarity
                    },
                    "title": {
                        "type": "text",
                        "analyzer": "ik_smart",
                    },
                    "source": {
                        "type": "text",
                    },
                    "pid": {
                        "type": "text",
                    },
                }
            }
        for key in self.search_fields.get_keys():
            mappings["properties"][key] = {
                "type": self.search_fields.get_field_type(key) or "text"
            }

        # Create the index with the specified settings and mappings
        if self.client.indices.exists(index=self.index_name):
            if self.drop_old:
                self.client.indices.delete(index=self.index_name)

        if not self.client.indices.exists(index=self.index_name):
            self.client.indices.create(
                index=self.index_name, mappings=mappings, settings=settings
            )

    def add_documents(
        self,
        documents: List[Document],
        refresh_indices: bool = True,
    ) -> List[str]:
        try:
            from elasticsearch.helpers import BulkIndexError, bulk
        except ImportError:
            raise ImportError(
                "Could not import elasticsearch python package. "
                "Please install it with `pip install elasticsearch`."
            )

        texts = [doc.page_content for doc in documents]
        metadatas = [doc.metadata for doc in documents]
        ids = [str(uuid.uuid4()) for _ in texts]
        requests = []

        for i, text in enumerate(texts):
            metadata = metadatas[i] if metadatas else {}

            request = {
                "_op_type": "index",
                "_index": self.index_name,
                "content": text,
                "title": metadata.get("title", ""),  # Index the title
                "_id": ids[i],
                "source": metadata.get("source"),
                "pid": metadata.get("pid"),
            }
            for filter in self.search_fields.get_keys():
                value = metadata.get(filter, "")
                if isinstance(value, list):
                    value = [v.strip() for v in value]
                elif value is not None:
                    value = value.strip()
                if value:
                    request[filter] = value
            requests.append(request)

        if len(requests) > 0:
            try:
                success, failed = bulk(
                    self.client,
                    requests,
                    stats_only=True,
                    refresh=refresh_indices,
                )
                logger.info(
                    f"Added {success} and failed to add {failed} texts to index"
                )

                return ids
            except BulkIndexError as e:
                logger.error(f"Error adding texts: {e}")
                firstError = e.errors[0].get("index", {}).get("error", {})
                logger.error(f"First error reason: {firstError.get('reason')}")
                raise e
        else:
            logger.info("No documents to add to index")
            return []

    def retrieve(
        self,
        query: str,
        k: int = 100,
        filter: Dict[str, Any] = {},
    ) -> List[Tuple[Document, float]]:
        assert self.client.indices.exists(index=self.index_name)
        start_time = time.time()
        query_dict = {
            "query": {
                "bool": {
                    "should": [
                        {
                            "match": {
                                "title": {
                                    "query": query,
                                    "boost": 2.0,
                                }
                            }
                        },
                        {
                            "match": {
                                "content": query,
                            },
                        },
                    ],
                    "must": [],
                }
            },
            "_source": True,
        }

        for field in filter:
            category_or_date = filter.get(field)
            if category_or_date:
                if isinstance(category_or_date, tuple):
                    query_dict["query"]["bool"]["must"].append(
                        {
                            "range": {
                                field: {
                                    "gte": category_or_date[0],
                                    "lte": category_or_date[1]
                                    if len(category_or_date) > 1
                                    else category_or_date[0],  # type: ignore
                                }
                            }
                        }
                    )
                else:
                    query_dict["query"]["bool"]["must"].append(
                        {"term": {field: category_or_date}}
                    )

        res = self.client.search(
            index=self.index_name,
            body=query_dict,
            size=k,
        )
        top_k_used = min(len(res["hits"]["hits"]), k)
        docs = []
        for id in range(top_k_used):
            _source = res["hits"]["hits"][id]["_source"]
            doc = Document(
                page_content=_source["content"],
                metadata={
                    "source": _source["source"],
                    "title": _source["title"],
                    "pid": _source["pid"],
                },
            )
            score = res["hits"]["hits"][id]["_score"]
            for field in self.search_fields.get_keys():
                if _source.get(field):
                    doc.metadata[field] = _source.get(field)
            docs.append((doc, score))
        retrieve_time_sec = time.time() - start_time
        logger.info(f"Keyword retrieve time: {retrieve_time_sec:.3f} sec.")
        logger.info(f"Retrieved {len(docs)} documents.")
        return docs

    def get_index_mappings(self):
        mapping = self.client.indices.get_mapping(index=self.index_name)

        # The mapping response structure can be quite nested, focusing on the 'properties' section
        properties = mapping[self.index_name]["mappings"]["properties"]

        # Function to recursively extract fields and types
        def extract_fields(fields_dict, parent_name=""):
            fields = {}
            for field_name, details in fields_dict.items():
                full_field_name = (
                    f"{parent_name}.{field_name}" if parent_name else field_name
                )
                if "properties" in details:
                    fields.update(
                        extract_fields(details["properties"], full_field_name)
                    )
                else:
                    fields[full_field_name] = details.get(
                        "type", "notype"
                    )  # Default 'notype' if no type is found
            return fields

        # Extract fields and types
        all_fields = extract_fields(properties)
        return all_fields

    def get_categories(self, field: str, k: int = 10):
        query = {
            "size": 0,  # No actual documents are needed, just the aggregation results
            "aggs": {
                "all_categories": {
                    "terms": {
                        "field": field,
                        "size": 1000,  # Adjust this value based on the expected number of unique categories
                    }
                }
            },
        }
        response = self.client.search(index=self.index_name, body=query)
        # Extract the aggregation results
        categories = response["aggregations"]["all_categories"]["buckets"]
        if k > 0:
            categories = categories[:k]
        res = [category["key"] for category in categories]
        return res

    def delete(
        self,
        ids: Optional[List[str]] = None,
        source_id: Optional[str] = None,
        source_url: Optional[str] = None,
        **kwargs: str,
    ):
        if ids:
            query = {"query": {"terms": {"uid": ids}}}
        elif source_id:
            query = {"query": {"match": {"source": source_id}}}
        elif source_url:
            query = {
                "query": {
                    "wildcard": {
                        "source": f"*{source_url}*"
                    }
                }
            }
        else:
            raise ValueError("Please provide either ids, source_id, or source_url to delete.")

        result = self.client.delete_by_query(index=self.index_name, body=query)
        deleted_count = result.get('deleted', 0)

        # Refresh the index to make the changes visible
        self.client.indices.refresh(index=self.index_name)

        logger.info(f"Deleted {deleted_count} documents with {'ids' if ids else 'source_id' if source_id else 'source_url'}: {ids or source_id or source_url}")

    def delete_all(self):
        self.client.indices.delete(index=self.index_name)



import logging
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import numpy as np
from langchain_core.documents import Document
from pymilvus import (
    Collection,
    CollectionSchema,
    DataType,
    FieldSchema,
    MilvusException,
    connections,
    utility,
)

from denser_retriever.embeddings import DenserEmbeddings
from denser_retriever.filter import FieldMapper
from denser_retriever.vectordb.base import DenserVectorDB

logger = logging.getLogger(__name__)

DEFAULT_MILVUS_CONNECTION = {
    "uri": "http://localhost:19530",
}


class MilvusDenserVectorDB(DenserVectorDB):
    def __init__(
        self,
        drop_old: Optional[bool] = False,
        auto_id: bool = False,
        connection_args: Optional[dict] = None,
        **args: Any,
    ):
        super().__init__(**args)
        self.drop_old = drop_old
        self.auto_id = auto_id
        self.connection_args = connection_args

    def _create_connection_alias(self, connection_args: dict) -> str:
        """Create the connection to the Milvus server."""

        # Grab the connection arguments that are used for checking existing connection
        host: str = connection_args.get("host", None)
        port: Union[str, int] = connection_args.get("port", None)
        address: str = connection_args.get("address", None)
        uri: str = connection_args.get("uri", None)
        user = connection_args.get("user", None)

        # Order of use is host/port, uri, address
        if host is not None and port is not None:
            given_address = str(host) + ":" + str(port)
        elif uri is not None:
            if uri.startswith("https://"):
                given_address = uri.split("https://")[1]
            elif uri.startswith("http://"):
                given_address = uri.split("http://")[1]
            else:
                given_address = uri  # Milvus lite
        elif address is not None:
            given_address = address
        else:
            given_address = None
            logger.debug("Missing standard address type for reuse attempt")

        # User defaults to empty string when getting connection info
        if user is not None:
            tmp_user = user
        else:
            tmp_user = ""

        # If a valid address was given, then check if a connection exists
        if given_address is not None:
            for con in connections.list_connections():
                addr = connections.get_connection_addr(con[0])
                if (
                    con[1]
                    and ("address" in addr)
                    and (addr["address"] == given_address)
                    and ("user" in addr)
                    and (addr["user"] == tmp_user)
                ):
                    logger.debug("Using previous connection: %s", con[0])
                    return con[0]

        # Generate a new connection if one doesn't exist
        alias = uuid4().hex
        try:
            connections.connect(alias=alias, **connection_args)
            logger.debug("Created new connection using: %s", alias)
            return alias
        except MilvusException as e:
            logger.error("Failed to create new connection using: %s", alias)
            raise e

    def create_index(
        self,
        index_name: str,
        embeddings: DenserEmbeddings,
        search_fields: List[str],
        **kwargs,
    ):
        """Create the index for the vector db."""
        self.index_name = index_name
        self.search_fields = FieldMapper(search_fields)
        self.embeddings = embeddings
        self.source_max_length = 500
        self.title_max_length = 500
        self.text_max_length = 8000
        self.field_max_length = 500

        self.connection_args = self.connection_args or DEFAULT_MILVUS_CONNECTION
        self.alias = self._create_connection_alias(self.connection_args)
        self.col: Optional[Collection] = None

        # Grab the existing collection if it exists
        if utility.has_collection(self.index_name, using=self.alias):
            self.col = Collection(
                self.index_name,
                using=self.alias,
            )
            if self.drop_old:
                self.col.drop()
                self.col = None

        # Either creates or references a collection. It does not remove records.
        self._create_collection()

    def _create_collection(self):
        fields = [
            FieldSchema(
                name="uid",
                dtype=DataType.VARCHAR,
                is_primary=True,
                auto_id=False,
                max_length=100,
            ),
            FieldSchema(
                name="source", dtype=DataType.VARCHAR, max_length=self.source_max_length
            ),
            FieldSchema(
                name="title", dtype=DataType.VARCHAR, max_length=self.title_max_length
            ),
            FieldSchema(
                name="text", dtype=DataType.VARCHAR, max_length=self.text_max_length
            ),
            FieldSchema(name="pid", dtype=DataType.VARCHAR, max_length=100),
            FieldSchema(
                name="embeddings", dtype=DataType.FLOAT_VECTOR, dim=self.embeddings.embedding_size
            ),
        ]
        for key in self.search_fields.get_keys():
            # both category and date type (unix timestamp) use INT64 type
            fields.append(
                FieldSchema(
                    name=key,
                    dtype=DataType.INT64,
                    max_length=self.field_max_length,
                ),
            )
        schema = CollectionSchema(fields=fields, description="Denser Vector DB")
        try:
            self.col = Collection(
                self.index_name,
                schema,
                consistency_level="Strong",
                using=self.alias,
            )
        except MilvusException as e:
            logger.error(
                "Failed to create collection: %s error: %s", self.index_name, e
            )
            raise e

    def add_documents(
        self,
        documents: List[Document],
        batch_size: int = 1000,
        **kwargs: Any,
    ) -> List[str]:
        """Add documents to the vector db.

        Args:
            documents (List[Document]): Documents to add to the vector db.

        Returns:
            List[str]: IDs of the added texts.
        """
        batch = []
        ids = [str(uuid4()) for _ in range(len(documents))]
        uid_list, sources, titles, texts, pid_list = [], [], [], [], []
        fields_list = [[] for _ in range(len(self.search_fields.get_keys()))]
        failed_batches = []  # To store information about failed batches
        for doc, id in zip(documents, ids):
            batch.append(
                (
                    doc.metadata.get("title", "")[: self.title_max_length - 10]
                    + " "
                    + doc.page_content[:2000]
                ).strip()
            )
            uid_list.append(id)
            sources.append(
                doc.metadata.get("source", "")[: self.source_max_length - 10]
            )
            titles.append(doc.metadata.get("title", "")[: self.title_max_length - 10])
            texts.append(doc.page_content[: self.text_max_length - 1000])  # buffer
            pid_list.append(doc.metadata.get("pid", "-1"))

            for i, field_original_key in enumerate(
                self.search_fields.get_original_keys()
            ):
                data = doc.metadata.get(field_original_key, -1)
                converted_data = self.search_fields.convert_for_storage(
                    {field_original_key: data}
                )
                fields_list[i].append(converted_data)

            if len(batch) == batch_size:
                embeddings = self.embeddings.embed_documents(batch)
                record = [
                    uid_list,
                    sources,
                    titles,
                    texts,
                    pid_list,
                    np.array(embeddings),
                ]
                record += fields_list

                try:
                    self.col.insert(record)
                except Exception as e:
                    logger.error(
                        f'Milvus index insert error at record {doc.metadata["pid"]} - {e}'
                    )

                self.col.flush()
                logger.info(f"Milvus vector DB ingesting {id}")

                batch = []
                uid_list, sources, titles, texts, pid_list = [], [], [], [], []
                fields_list = []

        if len(batch) > 0:
            embeddings = self.embeddings.embed_documents(batch)
            record = [
                uid_list,
                sources,
                titles,
                texts,
                pid_list,
                np.array(embeddings),
            ]
            record += fields_list
            try:
                self.col.insert(record)
            except Exception as e:
                logger.error(f"Milvus index insert error at record {id} - {e}")
                failed_batches.append(
                    {
                        "sources": sources,
                        "pids": pid_list,
                        "batch": batch,
                    }
                )
            self.col.flush()
            logger.info(f"Milvus vector DB ingesting {id}")

        index = {
            "index_type": "FLAT",
            "metric_type": "L2",
        }

        self.col.create_index("embeddings", index)
        self.col.load()
        return ids

    def similarity_search_with_score(
        self,
        query: str,
        k: int = 100,
        filter: Dict[str, Any] = {},
        **kwargs: Any,
    ) -> List[Tuple[Document, float]]:
        """Search for similar documents to the query.

        Args:
            query (str): Query text.
            k (int): Number of documents to return.
            param (Optional[dict]): Additional parameters for the search.
            expr (Optional[str]): Expression to filter the search.
            timeout (Optional[float]): Timeout for the search.

        Returns:
            List[Tuple[Document, float]]: List of tuples of documents and their similarity scores.
        """
        start_time = time.time()
        embeddings = self.embeddings.embed_query(query)
        query_embeddings = np.array(embeddings)
        embedding_time_sec = time.time() - start_time
        logger.info(f"Query embedding time: {embedding_time_sec:.3f} sec.")

        exprs = []
        for field in filter:
            original_key = filter.get(field)
            key = self.search_fields.get_key(original_key)
            type = self.search_fields.get_field_type(key)

            assert (
                original_key is not None
            ), f"Field {field} not found in the search fields."
            if type == "date":
                if len(original_key) == 2:
                    start_unix_time = int(
                        datetime.combine(
                            original_key[0], datetime.min.time()
                        ).timestamp()
                    )
                    end_unix_time = int(
                        datetime.combine(
                            original_key[1], datetime.min.time()
                        ).timestamp()
                    )
                    exprs.append(f"{key} >= {start_unix_time}")
                    exprs.append(f"{key} <= {end_unix_time}")
                else:
                    unix_time = int(
                        datetime.combine(
                            original_key[0], datetime.min.time()
                        ).timestamp()
                    )
                    exprs.append(f"{key} == {unix_time}")
            else:
                category_id = self.search_fields.get_key(original_key)
                if category_id is not None:
                    exprs.append(f"{key}=={category_id}")
        expr_str = " and ".join(exprs)
        search_params = {
            "metric_type": "L2",
            "params": {"nprobe": 10},
        }
        output_fields = [
            "source",
            "title",
            "text",
            "pid",
            "uid",
        ] + self.search_fields.get_keys()

        start_time = time.time()
        result = self.col.search(
            data=query_embeddings,
            anns_field="embeddings",
            param=search_params,
            limit=k,
            expr=expr_str,
            output_fields=output_fields,
        )
        retrieve_time_sec = time.time() - start_time
        logger.info(f"Vector DB retrieve time: {retrieve_time_sec:.3f} sec.")
        logger.info(f"Retrieved {len(result[0])} documents.")

        top_k_used = min(len(result[0]), self.top_k)  # type: ignore

        ret = []
        for id in range(top_k_used):
            assert len(result) == 1  # type: ignore
            hit = result[0][id]  # type: ignore
            doc = Document(page_content=hit.entity.text, metadata={})
            doc.metadata = {
                "id": hit.entity.uid,
                "source": hit.entity.source,
                "text": hit.entity.text,
                "title": hit.entity.title,
                "pid": hit.entity.pid,
            }
            score = -hit.entity.distance

            for field in self.search_fields.get_keys():
                original_value = self.search_fields.convert_to_original(
                    {field: hit.entity.get(field)}
                )
                doc.metadata[field] = original_value
            #     cat_id_or_unix_time = hit.entity.get(key)
            #     type = self.search_fields.get_field_type(field)
            #     if type == "date":
            #         date = datetime.utcfromtimestamp(cat_id_or_unix_time).strftime(
            #             "%Y-%m-%d"
            #         )
            #         doc.metadata[field] = date
            #     else:
            #         doc.metadata[field] = cat_id_or_unix_time
            pair = (doc, score)
            ret.append(pair)
        return ret

    def filter_expression(
        self,
        filter_dict: Dict[str, Any],
    ) -> Any:
        """Generate a Milvus expression from a filter dictionary."""
        expressions = []
        for key, value in filter_dict.items():
            if value is None:
                continue
            if isinstance(value, tuple) and len(value) == 2:
                start, end = value
                expressions.append(f"{key} >= '{start}' and {key} <= '{end}'")
            else:
                expressions.append(f"{key} == '{value}'")
        return " and ".join(expressions)

    def delete(
        self,
        ids: Optional[List[str]] = None,
        source_id: Optional[str] = None,
        source_url: Optional[str] = None,
    ):
        """Delete documents from the vector db.

        Args:
            ids (Optional[List[str]]): IDs of the documents to delete.
            expr (Optional[str]): Expression to filter the deletion.
        """
        # self.col.delete(ids=ids, **kwargs)
        if isinstance(ids, list) and len(ids) > 0:
            if source_id is not None:
                logger.warning(
                    "Both ids and source_id are provided. " "Ignore source_id and delete by ids."
                )
            expr = f"uid in {ids}"
            self.col.delete(expr=expr)
        elif source_id:
            self.col.delete(expr=f"source == '{source_id}'")
        elif source_url:
            self.col.delete(expr=f"source like '{source_url}%'")
        else:
            raise ValueError("No ids or source_id provided for deletion")

    def delete_all(self):
        """Delete all documents from the vector db."""
        if self.col:
            self.col.delete(expr="uid != ''")
